//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#ifndef __Coinc2__
#define __Coinc2__

#include "GwollumPlot.h"
#include "ReadTriggers.h"
#include "MakeTriggers.h"
#include "TriggerSelect.h"

using namespace std;

/**
 * Set 2 trigger sets in coincidence.
 * The coincidence algorithm includes a time and a frequency coincidence scheme.
 * \author    Florent Robinet
 */
class Coinc2 : public GwollumPlot {

 public:

  /**
   * @name Constructors and destructors
   @{
  */
  /**
   * Constructor of the Coinc2 class.
   * Sets parameters to default:
   * - coincidence time distance = 0.0s (see SetCoincDt())
   * - coincidence frequency distance = 0.0Hz (see SetCoincDf())
   * - coincidence frequency asymmetry = 1.0 (see SetCoincAf())
   * - no time offset applied to the clusters (see SetTriggerSamples())
   *
   * Two input triggger samples must be given. They are indexed 0 and 1. Triggers must be clustered before using this class! Triggers must not be changed when using the Coinc2 class methods. Use SetTriggerSamples() to define the input trigger samples.
   * @param aVerbose verbosity level
   */
  Coinc2(const int aVerbose=0);

  /**
   * Constructor of the Coinc2 class.
   * Sets parameters to default:
   * - coincidence time distance = 0.0s (see SetCoincDt())
   * - coincidence frequency distance = 0.0Hz (see SetCoincDf())
   * - coincidence frequency asymmetry = 1.0 (see SetCoincAf())
   * - no time offset applied to the cluster (see SetTriggerSample())
   *
   * Two input triggger samples must be given. They are indexed 0 and 1. Triggers must be clustered before using this class! Triggers must not be changed when using the Coinc2 class methods. If needed, use SetTriggerSamples() to change the input trigger samples defined in this constructor.
   * @param aTriggerSample0 first input trigger sample
   * @param aTriggerSample1 second input trigger sample
   * @param aVerbose verbosity level
   */
  Coinc2(ReadTriggers *aTriggerSample0, ReadTriggers *aTriggerSample1, const int aVerbose=0);

  /**
   * Destructor of the Coinc2 class.
   */
  virtual ~Coinc2(void);
  /**
     @}
  */

  /**
   * Sets trigger samples.
   * This function is used to define the input trigger samples. This function is also to be used after any change performed in the trigger sample (with ReadTriggers methods). Finally, this function can be used to apply a time offset to the clusters.
   * - coinc segments are defined as the intersection of both trigger samples (using time offsets and user time segements).
   * - previous coinc events are deleted from memory
   * - clusters are tagged: 1 if the peak time is inside a coinc segment, -1 otherwise (the cluster is discarded by the methods of this class).
   * - coinc monitors are reset
   *
   * IMPORTANT1: input triggers must be clustered first!
   * IMPORTANT2: by default, if clusters are tagged with -1 value before calling this function, they will not be considered at all, even if inside a coinc segment. This behavior can be canceled with the aIgnoreMinusOnes argument.
  * @param aTriggerSample0 pointer to the first trigger object
   * @param aTriggerSample1 pointer to the second trigger object
   * @param aTimeOffset0 time offset to apply to the clusters of the first sample
   * @param aTimeOffset1 time offset to apply to the clusters of the second sample
   * @param aValidSegments list of segments where to consider clusters (before time offset!)
   * @param aIgnoreMinusOnes if this parameter is set to true, clusters tagged with -1 are ignored
   */
  bool SetTriggerSamples(ReadTriggers *aTriggerSample0, ReadTriggers *aTriggerSample1, const double aTimeOffset0=0.0, const double aTimeOffset1=0.0, Segments *aValidSegments=NULL, const bool aIgnoreMinusOnes=true);

  /**
   * Runs the reference coincidence algorithm.
   * Returns the number of coinc events. Returns -1 if this function fails.
   * 2 clusters are said to coincide if 3 conditions are verified:
   * - the time distance between the cluster time limits is smaller than the coinc parameter set with SetCoincDt() (note that the time offset defined with SetTriggerSample() is applied).
   * - the frequency distance between the cluster frequency limits is smaller than the coinc parameter set with SetCoincDf().
   * - the frequency asymmetry between the cluster peak frequency is smaller or equal to the coinc parameter set with SetCoincAf().
   *
   * Coinc events can be tagged (see SetCoincTag()), by default, all coincs are tagged to true.
   */
  int MakeCoinc(void);

  /**
   * Runs the 0-type coincidence algorithm.
   * Returns the number of coinc events. Returns -1 if this function fails. For this coincidence algorithm, a time window is defined around the peak time of cluster (0). By default the time window is symetric around the peak time and the window width is given by the time parameter set with SetCoincDt(). However, if the window size exceeds the size of cluster (0), the window is ajusted back to the cluster time edges.
   * 2 clusters (0) and (1) are said to coincide if the cluster peak time of (1) is inside the time window defined with (0) (note that the time offset defined with SetTriggerSample() is applied).
   *
   * Note: there is no frequency coicidence.
   * Coinc events can be tagged (see SetCoincTag()), by default, all coincs are tagged to true.
   */
  int MakeCoinc0(void);

  /**
   * Prints coinc events.
   */
  void PrintCoinc(void);

  /**
   * Writes coinc events on disk.
   * Two trigger structures (See MakeTriggers) are saved in two separate root directories (clusters_0 and clusters_1). The trigger structures contain parameters for clusters involved in a coinc event.
   * Only coinc with a tag set to true are saved, see SetCoincTag().
   * Additionally, a tree named coinc2param is saved. It contains a single entry with the parameters used for the coincidence (time and frequency windows, time offsets etc.).
   * @param aOutDir output directory (must exist)
   * @param aOutFileName output file name (do not provide an extension)
   */
  bool WriteCoinc(const string aOutDir, const string aOutFileName);

  /**
   * Sets the maximum time distance to define a coinc.
   * @param aDt maximum time distance value [s]
   */
  inline void SetCoincDt(const double aDt){ coinc_dt=aDt; return; };

  /**
   * Sets the maximum frequency distance to define a coinc.
   * @param aDf maximum frequency distance value [Hz]
   */
  inline void SetCoincDf(const double aDf){ coinc_df=aDf; return; };

  /**
   * Sets the maximum frequency asymmetry to define a coinc.
   * @param aDf maximum frequency asymmetry value [between 0 and 1]
   */
  inline void SetCoincAf(const double aAf){ coinc_af=aAf; return; };

  /**
   * Returns the current number of coinc events.
   */
  inline int GetNCoinc(void){ return (int)CoC[0].size(); };

  /**
   * Returns the cluster index involved in a given coinc event.
   * IMPORTANT: the validity of indexes is not checked.
   * @param aSampleIndex trigger sample index: 0 or 1
   * @param aCoincIndex coinc index
   */
  inline int GetClusterIndex(const int aSampleIndex, const int aCoincIndex){ 
    return CoC[aSampleIndex][aCoincIndex];
  };

  /**
   * Sets a new tag to a given coinc.
   * IMPORTANT: the validity of indexes is not checked.
  * @param aCoincIndex coinc index
   * @param aNewTag new tag value
   */
  inline void SetCoincTag(const int aCoincIndex, const bool aNewTag){ 
    CoTag[aCoincIndex]=aNewTag;
  };

  /**
   * Sets a new tag for all coincs.
   * @param aNewTag new tag value
   */
  inline void SetCoincTag(const bool aNewTag){ 
    for(int c=0; c<(int)CoC[0].size(); c++) CoTag[c]=aNewTag;
  };

  /**
   * Gets the tag value of a given coinc.
   * IMPORTANT: the validity of indexes is not checked.
   * @param aCoincIndex coinc index
   */
  inline bool GetCoincTag(const int aCoincIndex){ return CoTag[aCoincIndex]; };

  /**
   * Gets the number of active clusters.
   * IMPORTANT: the validity of indexes is not checked.
   * @param aSampleIndex trigger sample index: 0 or 1
   */
  inline int GetNActiveClusters(const int aSampleIndex){ 
    return nactive[aSampleIndex];
  };

  /**
   * Returns the time offset of a trigger sample.
   * IMPORTANT: the validity of indexes is not checked.
   * @param aSampleIndex trigger sample index: 0 or 1
   */
  inline double GetTimeOffset(const int aSampleIndex){ 
    return toffset[aSampleIndex];
  };

  /**
   * Returns the time offset of a trigger sample.
   * @param aSampleIndex trigger sample index: 0 or 1
   */
  inline double GetCoincDt(void){ return coinc_dt; };

  /**
   * Returns the live time of coinc segments.
   */
  inline double GetCoincLiveTime(void){ return CoSeg->GetLiveTime(); };

  /**
   * Sets a minimum frequency for a given cluster sample.
   * @param aSampleIndex trigger sample index: 0 or 1
   * @param aFreqMin minimum frequency [Hz]
   */
  inline bool SetFrequencyMin(const int aSampleIndex, const double aFreqMin){ 
    if(aSampleIndex < 0 || aSampleIndex > 1) return false;
    freqmin[aSampleIndex]=aFreqMin;
    return true;
  };

  /**
   * Sets a maximum frequency for a given cluster sample.
   * @param aSampleIndex trigger sample index: 0 or 1
   * @param aFreqMax maximum frequency [Hz]
   */
  inline bool SetFrequencyMax(const int aSampleIndex, const double aFreqMax){ 
    if(aSampleIndex < 0 || aSampleIndex > 1) return false;
    freqmax[aSampleIndex]=aFreqMax;
    return true;
  };


  /**
   * Fills monitors.
   * When called after performing a coincidence, this function fills 2-dimensional (cluster SNR/frequency) histograms, called monitors. There are 3 types of monitors, they are filled with:
   * - the number of valid clusters
   * - the number of valid clusters involved in a coinc
   * - the number of valid clusters not involved in a coinc
   *
   * If requested, clusters can be weighted by 1.0/(cluster duration)
   * @param aWeighted flag to use weighted monitors
   */
  //bool MakeMonitors(const bool aWeighted = false);

  /**
   * Writes monitors on disk (root file).
   * @param aOutDir output directory (must exist)
   * @param aOutFileName output file name (do not provide an extension)
   */
  //bool WriteMonitors(const string aOutDir, const string aOutFileName);

  bool MakeComparators(const string aTriggerId0, const string aTriggerId1);
  bool PrintComparator(const string aComparatorName);

 protected:

  // GENERAL
  int verbosity;              ///< verbosity level

  // INPUT
  ReadTriggers *triggers[2];  ///< trigger samples
  double toffset[2];          ///< time offset
  int nactive[2];             ///< number of active clusters
  double freqmin[2];          /// frequency min selection
  double freqmax[2];          /// frequency max selection

  // COINC
  vector <int> CoC[2];        ///< coinc cluster index
  bool *CoTag;                ///< coinc tag
  Segments *CoSeg;            ///< coinc segments
  TTree *CoTree;              ///< coinc tree

  // PARAMETERS
  double coinc_dt;            ///< coinc time distance
  double coinc_df;            ///< coinc frequency distance
  double coinc_af;            ///< coinc frequency asymmetry

  // MONITORS
  TH2D *h_ct[2];               ///< number of active clusters
  TH2D *h_ctcum[2];            ///< cumulative number of active clusters
  TH2D *h_ct_coinc[2];         ///< number of clusters in a coinc
  TH2D *h_ctcum_coinc[2];      ///< cumulative number of clusters in a coinc

  // COMPARATORS
  TH2D *hc_asnr;              ///< SNR asymmetry vs SNR
  TH2D *hc_snrsnr;            ///< SNR vs snr
  TH2D *hc_af;                ///< frequency asymmetry vs SNR
  TH2D *hc_ff;                ///< frequency vs frequency


 private:
  
  void SelectClusters(const int aSampleIndex, const bool aIgnoreMinusOnes=true);  ///< select clusters in coinc seg

  ClassDef(Coinc2,0)  
};

#endif


