//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "InjGen.h"

ClassImp(InjGen)


/////////////////////////////////////////////////////////////////////////
InjGen::InjGen(Segments *aInSegments, const string aInjType, const int aVerbose) : InjTre (aInjType,aVerbose){ 
/////////////////////////////////////////////////////////////////////////

  // parameters
  InSegments=aInSegments;
  InjSpacing=100;

  // random generator
  rand = new TRandom3();
  rand->SetSeed(0); // random seed

  // Set default modes
  SetAmplitudeDistribution();
  SetAmplitudeRange();
  SetSigmaDistribution();
  SetSigmaRange();
  SetFrequencyDistribution();
  SetFrequencyRange();
  SetEccentricityDistribution();
  SetEccentricityRange();
  SetPolarizationDistribution();
  SetPolarizationRange();
}

/////////////////////////////////////////////////////////////////////////
InjGen::~InjGen(void){
/////////////////////////////////////////////////////////////////////////
  delete rand;
}

/////////////////////////////////////////////////////////////////////////
int InjGen::GenerateInjections(void){
/////////////////////////////////////////////////////////////////////////

  // no livetime
  if(!InSegments->GetLiveTime()) return 0;

  // init timing
  double gps=InSegments->GetStart(0)+rand->Uniform(InjSpacing);
  int current_seg=0;
  
  // loop over injections
  while(gps<InSegments->GetEnd(InSegments->GetNsegments()-1)){

    // injection inside segment -> save it
    if(InSegments->IsInsideSegment(gps,current_seg)){
      
      if(!GetInjectionType().compare("Gauss"))
	AddGauss(rand->Uniform(0,2*TMath::Pi()), // ra
		 TMath::ASin(rand->Uniform(-1,1)), //dec
		 gps,
		 GetEccentricity(),
		 GetPolarization(),
		 GetAmplitude(),
		 GetSigma()
		 );

      if(!GetInjectionType().compare("SineGauss"))
	AddSineGauss(rand->Uniform(0,2*TMath::Pi()), // ra
		     TMath::ASin(rand->Uniform(-1,1)), //dec
		     gps,
		     GetEccentricity(),
		     GetPolarization(),
		     GetAmplitude(),
		     GetSigma(),
		     GetFrequency()
		     );
    }
    
    // injection outside segment -> move to next segment
    else{
      if(current_seg+1<InSegments->GetNsegments()){
	gps=InSegments->GetStart(current_seg+1)+rand->Uniform(InjSpacing);
	continue;
      }
      else break;
    }

    // increment GPS
    gps+=InjSpacing;
  }
  
  return GetNInjections();
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetAmplitudeDistribution(const string aDistribution){
/////////////////////////////////////////////////////////////////////////
  if(!aDistribution.compare("UNIFORM"))  ampDistribution=0;
  else if(!aDistribution.compare("LOG")) ampDistribution=1;
  else{
    cerr<<"InjGen::SetAmplitudeDistribution: the amplitude distribution "<<aDistribution<<" is not supported"<<endl;
    return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetAmplitudeRange(const double aMin, const double aMax){
/////////////////////////////////////////////////////////////////////////
  if(aMax<aMin){
    cerr<<"InjGen::SetAmplitudeRange: the amplitude range "<<aMin<<"-"<<aMax<<" is incorrect"<<endl;
    return false;
  }
  ampMin=aMin;
  ampMax=aMax;
  return true;
}

/////////////////////////////////////////////////////////////////////////
double InjGen::GetAmplitude(void){
/////////////////////////////////////////////////////////////////////////
  if(ampDistribution==0)      return rand->Uniform(ampMin,ampMax);
  else if(ampDistribution==1) return ampMin*pow(ampMax/ampMin,rand->Uniform());
  else                        return 0.0;
  return 0.0;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetSigmaDistribution(const string aDistribution){
/////////////////////////////////////////////////////////////////////////
  if(!aDistribution.compare("UNIFORM"))  sigmaDistribution=0;
  else if(!aDistribution.compare("LOG")) sigmaDistribution=1;
  else{
    cerr<<"InjGen::SetSigmaDistribution: the sigma distribution "<<aDistribution<<" is not supported"<<endl;
    return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetSigmaRange(const double aMin, const double aMax){
/////////////////////////////////////////////////////////////////////////
  if(aMax<aMin){
    cerr<<"InjGen::SetSigmaRange: the sigma range "<<aMin<<"-"<<aMax<<" is incorrect"<<endl;
    return false;
  }
  sigmaMin=aMin;
  sigmaMax=aMax;
  return true;
}

/////////////////////////////////////////////////////////////////////////
double InjGen::GetSigma(void){
/////////////////////////////////////////////////////////////////////////
  if(sigmaDistribution==0)      return rand->Uniform(sigmaMin,sigmaMax);
  else if(sigmaDistribution==1) return sigmaMin*pow(sigmaMax/sigmaMin,rand->Uniform());
  else                  return 0.0;
  return 0.0;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetFrequencyDistribution(const string aDistribution){
/////////////////////////////////////////////////////////////////////////
  if(!aDistribution.compare("UNIFORM"))  f0Distribution=0;
  else if(!aDistribution.compare("LOG")) f0Distribution=1;
  else{
    cerr<<"InjGen::SetFrequencyDistribution: the frequency distribution "<<aDistribution<<" is not supported"<<endl;
    return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetFrequencyRange(const double aMin, const double aMax){
/////////////////////////////////////////////////////////////////////////
  if(aMax<aMin||aMin<=0.0){
    cerr<<"InjGen::SetFrequencyRange: the frequency range "<<aMin<<"-"<<aMax<<" is incorrect"<<endl;
    return false;
  }
  f0Min=aMin;
  f0Max=aMax;
  return true;
}

/////////////////////////////////////////////////////////////////////////
double InjGen::GetFrequency(void){
/////////////////////////////////////////////////////////////////////////
  if(f0Distribution==0)       return rand->Uniform(f0Min,f0Max);
  else if(f0Distribution==1)  return f0Min*pow(f0Max/f0Min,rand->Uniform());
  else                        return 0.0;
  return 0.0;
}


/////////////////////////////////////////////////////////////////////////
bool InjGen::SetEccentricityDistribution(const string aDistribution){
/////////////////////////////////////////////////////////////////////////
  if(!aDistribution.compare("UNIFORM"))  eccDistribution=0;
  else if(!aDistribution.compare("LOG")) eccDistribution=1;
  else{
    cerr<<"InjGen::SetEccentricityDistribution: the eccentricity distribution "<<aDistribution<<" is not supported"<<endl;
    return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetEccentricityRange(const double aMin, const double aMax){
/////////////////////////////////////////////////////////////////////////
  if(aMax<aMin||aMin<0.0||aMax>1.0){
    cerr<<"InjGen::SetEccentricityRange: the eccentricity range "<<aMin<<"-"<<aMax<<" is incorrect"<<endl;
    return false;
  }
  eccMin=aMin;
  eccMax=aMax;
  return true;
}

/////////////////////////////////////////////////////////////////////////
double InjGen::GetEccentricity(void){
/////////////////////////////////////////////////////////////////////////
  if(eccDistribution==0)       return rand->Uniform(eccMin,eccMax);
  else if(eccDistribution==1)  return eccMin*pow(eccMax/eccMin,rand->Uniform());
  else                         return 0.0;
  return 0.0;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetPolarizationDistribution(const string aDistribution){
/////////////////////////////////////////////////////////////////////////
  if(!aDistribution.compare("UNIFORM"))  psiDistribution=0;
  else if(!aDistribution.compare("LOG")) psiDistribution=1;
  else{
    cerr<<"InjGen::SetPolarizationDistribution: the polarization distribution "<<aDistribution<<" is not supported"<<endl;
    return false;
  }
  return true;
}

/////////////////////////////////////////////////////////////////////////
bool InjGen::SetPolarizationRange(const double aMin, const double aMax){
/////////////////////////////////////////////////////////////////////////
  if(aMax<aMin||aMin<0.0||aMax>2*TMath::Pi()){
    cerr<<"InjGen::SetPolarizationRange: the polarization range "<<aMin<<"-"<<aMax<<" is incorrect"<<endl;
    return false;
  }
  psiMin=aMin;
  psiMax=aMax;
  return true;
}

/////////////////////////////////////////////////////////////////////////
double InjGen::GetPolarization(void){
/////////////////////////////////////////////////////////////////////////
  if(psiDistribution==0)       return rand->Uniform(psiMin,psiMax);
  else if(psiDistribution==1)  return psiMin*pow(psiMax/psiMin,rand->Uniform());
  else                         return 0.0;
  return 0.0;
}



