//////////////////////////////////////////////////////////////////////////////
//  Author : florent robinet (LAL - Orsay): robinet@lal.in2p3.fr
//////////////////////////////////////////////////////////////////////////////
#include "Triggers.h"

ClassImp(Triggers)

////////////////////////////////////////////////////////////////////////////////////
Triggers::Triggers(const int aVerbose){ 
////////////////////////////////////////////////////////////////////////////////////

  // parameters
  Verbose     = aVerbose;
  readmode    = false;      // building mode
  Cdelta_t    = 0.1;        // 100ms time clustering
  Csizemin    = -1;         // no size selection
  cl_snr_thr  = -1;         // no SNR threshold for clustering
  Cmeandur    = 0.0;        // cluster mean duration
  Cmaxdur     = 0.0;        // cluster tot duration
  Ctotdur     = 0.0;        // cluster max duration
  cl_tag      = new int [0];// cluster tag
  Cph         = -100.0;   // when phase is missing (to be removed later)
  Tph         = -100.0;   // when phase is missing (to be removed later)

  // random id
  srand(time(NULL));
  randid = rand();
  ostringstream tmpstream;
  tmpstream<<randid;
  srandid = tmpstream.str();
  
  // init cluster tree
  Ctree = new TTree(("clusters_"+srandid).c_str(), "clusters");
  Ctree->Branch("time",       &Ctime,      "time/D");
  Ctree->Branch("tstart",     &Ctstart,    "tstart/D");
  Ctree->Branch("tend",       &Ctend,      "tend/D");
  Ctree->Branch("frequency",  &Cfreq,      "frequency/D");
  Ctree->Branch("fstart",     &Cfstart,    "fstart/D");
  Ctree->Branch("fend",       &Cfend,      "fend/D");
  Ctree->Branch("snr",        &Csnr,       "snr/D");
  Ctree->Branch("amplitude",  &Camp,       "amplitude/D");
  Ctree->Branch("q",          &Cq,         "q/D");
  Ctree->Branch("phase",      &Cph,        "phase/D");
  Ctree->Branch("firstentry", &Cfirstentry,"firstentry/L");
  Ctree->Branch("size",       &Csize,      "size/L");

  // init trigger trees
  Ttree_make = new TTree(("triggers_"+srandid).c_str(), "triggers");
  Ttree_make->Branch("time",     &Ttime,  "time/D");
  Ttree_make->Branch("frequency",&Tfreq,  "frequency/D");
  Ttree_make->Branch("q",        &Tq,     "q/D");
  Ttree_make->Branch("snr",      &Tsnr,   "snr/D");
  Ttree_make->Branch("tstart",   &Ttstart,"tstart/D");
  Ttree_make->Branch("tend",     &Ttend,  "tend/D");
  Ttree_make->Branch("fstart",   &Tfstart,"fstart/D");
  Ttree_make->Branch("fend",     &Tfend,  "fend/D");
  Ttree_make->Branch("amplitude",&Tamp,   "amplitude/D");
  Ttree_make->Branch("phase",    &Tph,    "phase/D");
  Ttree_make->Branch("tstart_ms",&Ttstart_ms,"tstart_ms/I");
  Ttree=Ttree_make;
}

////////////////////////////////////////////////////////////////////////////////////
Triggers::~Triggers(void){
////////////////////////////////////////////////////////////////////////////////////
  if(Verbose>1) cout<<"Triggers::~Triggers"<<endl;
  ResetTriggers();
  delete cl_tag;
  delete Ctree;
  delete Ttree_make;
}

////////////////////////////////////////////////////////////////////////////////////
void Triggers::ResetTriggers(void){ 
////////////////////////////////////////////////////////////////////////////////////
  ResetClusters();
  Ttree_make->Reset();
  if(readmode){
    readmode = false;      // back to building mode
    Ttree=Ttree_make;
    delete Ttree_read;
  }
  Tph=-100.0;// FIXME: should be removed
  return;
}

////////////////////////////////////////////////////////////////////////////////////
bool Triggers::ReadTriggerFiles(const string aPattern, const string aDirectory){ 
////////////////////////////////////////////////////////////////////////////////////
  
  // reset triggers
  ResetTriggers();

  // switch to reading mode
  readmode = true;

  // extract trigger trees
  if(Verbose) cout<<"Triggers::ReadTriggerFiles: loading triggers in \""<<aPattern<<"\""<<endl;
  if(aDirectory.compare("")) Ttree_read = new TChain((aDirectory+"/triggers").c_str());
  else  Ttree_read = new TChain("triggers");

  // get list of patterns and load it
  vector <string> patlist = SplitString(aPattern,' ');
  for(int p=0; p<(int)patlist.size(); p++) Ttree_read->Add(patlist[p].c_str());
  Ttree_read->SetName(("triggers_"+srandid).c_str());
  patlist.clear();
  
  // get tree offsets
  Toffsets=Ttree_read->GetTreeOffset();
  
  // attach branches
  if(Ttree_read->SetBranchAddress("time",&Ttime)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'time' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("frequency",&Tfreq)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'frequency' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("q",&Tq)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'q' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("snr",&Tsnr)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'snr' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("tstart",&Ttstart)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'tstart' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("tend",&Ttend)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'tend' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("fstart",&Tfstart)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'fstart' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("fend",&Tfend)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'fend' branch"<<endl;
    return false;
  }
  if(Ttree_read->SetBranchAddress("amplitude", &Tamp)<0){
    cerr<<"Triggers::ReadTriggerFiles: missing or corrupted 'amplitude' branch"<<endl;
    return false;
  }
  if(Ttree_read->GetBranchStatus("phase")) Ttree_read->SetBranchAddress("phase", &Tph);// FIXME: should be required

  Ttree=Ttree_read;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Triggers::AddTrigger(const double aTime,   const double aFrequency,
			  const double aSNR,    const double aQ,
			  const double aTstart, const double aTend,
			  const double aFstart, const double aFend,
			  const double aAmplitude, const double aPhase){
////////////////////////////////////////////////////////////////////////////////////
  if(readmode){
    cerr<<"Triggers::AddTrigger: this function is not available in a reading mode"<<endl;
    return false;
  }

  // get trigger variables
  Ttime=floor(aTime*1000000.0)/1000000.0;
  Tfreq=aFrequency;
  Tsnr=aSNR;
  Tq=aQ;
  Tamp=aAmplitude;
  Tph=aPhase;
  Ttstart=TMath::Min(floor(aTstart*1000000.0)/1000000.0,Ttime); 
  Ttend=TMath::Max(ceil(aTend*1000000.0)/1000000.0,Ttime); 
  Tfstart=TMath::Min(aFstart,Tfreq);
  Tfend=TMath::Max(aFend,Tfreq);
  Ttstart_ms=(int)floor((aTstart-(int)aTstart)*1000000+0.5);

  // print info
  if(Verbose>2) cout<<"Add trigger: time="<<Ttime<<", "
		     <<"frequency="<<Tfreq<<", "
		     <<"SNR="<<Tsnr<<", "
		     <<"q="<<Tq<<", "
		     <<"amplitude="<<Tamp<<", "
		     <<"phase="<<Tph<<", "
		     <<"tstart="<<Ttstart<<", "
		     <<"tend="<<Ttend<<", "
		     <<"fstart="<<Tfstart<<", "
		     <<"fend="<<Tfend<<endl;
  // fill tree
  if(Ttree->Fill()<0) return false;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
bool Triggers::SortTriggers(void){
////////////////////////////////////////////////////////////////////////////////////

  if(readmode){
    cerr<<"Triggers::SortTriggers: this function is not available in a reading mode"<<endl;
    return false;
  }
  if(Verbose) cout<<"Triggers::SortTriggers: sorting triggers"<<endl;
  if(!Ttree_make->GetEntries()) return true;

  // build new index
  if(Ttree_make->BuildIndex("tstart","tstart_ms")<0){
    cerr<<"Triggers::SortTriggers: the tstart indexing failed"<<endl;
    return false;
  }
   
  TTreeIndex *index = (TTreeIndex*)Ttree_make->GetTreeIndex();
  Long64_t N = index->GetN();
  TTree *sortedtree = Ttree_make->CloneTree(0);

  // fill sorted tree
  for(Long64_t t=0; t<N; t++){
    Ttree_make->GetEntry(index->GetIndex()[t]);
    sortedtree->Fill();
  }
  delete index;
  delete Ttree_make;

  // re-address trees
  sortedtree->SetBranchAddress("time",     &Ttime);
  sortedtree->SetBranchAddress("frequency",&Tfreq);
  sortedtree->SetBranchAddress("q",        &Tq);
  sortedtree->SetBranchAddress("snr",      &Tsnr);
  sortedtree->SetBranchAddress("tstart",   &Ttstart);
  sortedtree->SetBranchAddress("tend",     &Ttend);
  sortedtree->SetBranchAddress("fstart",   &Tfstart);
  sortedtree->SetBranchAddress("fend",     &Tfend);
  sortedtree->SetBranchAddress("amplitude",&Tamp);
  sortedtree->SetBranchAddress("phase",    &Tph);
  sortedtree->SetBranchAddress("tstart_ms",&Ttstart_ms);
  Ttree_make=sortedtree;
  Ttree=Ttree_make;

  return true;
}

////////////////////////////////////////////////////////////////////////////////////
Segments* Triggers::GetTriggerSegments(TH1D *aThr, const double aInfValue){ 
////////////////////////////////////////////////////////////////////////////////////

  Segments *S = new Segments();

  if(aThr==NULL){
    for(int t=0; t<(int)Ttree->GetEntries(); t++){
      Ttree->GetEntry(t);
      S->AddSegment(Ttstart,Ttend);
    }
  }
  else{
    double fmin = aThr->GetXaxis()->GetXmin();
    double fmax = aThr->GetXaxis()->GetXmax();
    int binn;
    for(int t=0; t<(int)Ttree->GetEntries(); t++){
      Ttree->GetEntry(t);
      if(Tfreq<fmin) continue;
      if(Tfreq>=fmax) continue;
      binn = aThr->GetXaxis()->FindBin(Tfreq); // get frequency bin
      if(aThr->GetBinContent(binn)>=aInfValue) continue; // infinity threshold
      if(Tsnr<aThr->GetBinContent(binn)) continue; // below threshold
      S->AddSegment(Ttstart,Ttend);
    }
  }

  return S;
}

////////////////////////////////////////////////////////////////////////////////////
string Triggers::GetTriggerFile(const int aTr){ 
////////////////////////////////////////////////////////////////////////////////////
  if(!readmode){
    cerr<<"Triggers::GetTriggerTime: this function is not available in a building-mode"<<endl;
    return "none";
  }
  if(Ttree_read->GetEntry(aTr)<=0){
    cerr<<"Triggers::GetTriggerTime: the trigger index "<<aTr<<" could not be found."<<endl;
    return "none";
  }
  return Ttree_read->GetFile()->GetName();
}

////////////////////////////////////////////////////////////////////////////////////
bool Triggers::SetTriggerBranchStatus(const string aBname, const bool aStatus){ 
////////////////////////////////////////////////////////////////////////////////////
  UInt_t found;
  Ttree->SetBranchStatus(aBname.c_str(), aStatus, &found);
  return (bool)found;
}

////////////////////////////////////////////////////////////////////////////////////
int Triggers::GetTriggerIndex(const double aTime){ 
////////////////////////////////////////////////////////////////////////////////////
  // loop over triggers
  for(int t=0; t<Ttree->GetEntries(); t++){
    Ttree->GetEntry(t);
    if(aTime>=Ttstart&&aTime<Ttend) return t;
  }
  return -1; // not found

  // FIXME: optimize this for reading-mode
}

